/**
 * Created by IntelliJ IDEA.
 * User: Kilho Shin
 * Date: 11/02/21
 * Time: 11:40
 * To change this template use File | Settings | File Templates.

  This file is part of Mapping Kernels for Trees.

   Mapping Kernels for Trees is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   Mapping Kernels for Trees is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Mapping Kernels for Trees.  If not, see <http://www.gnu.org/licenses/>.

   For more details about this code, please look at the companion paper
   "Mapping Kernels for Trees", K. Shin, M.Cuturi, T.Kuboyama, Proceedings of
   the International Conference on Machine Learning 2011.

 */

import scala.collection.mutable.Map

class BnEval(val x: Double, val y: Double) {
  val xpow: Map[Int, Double] = Map(0->1)
  val ypow: Map[Int, Double] = Map(0->1)

  def x_raised_to(i: Int): Double = {
    if(!xpow.isDefinedAt(i))
      xpow += i -> x_raised_to(i-1)*x
    xpow(i)
  }

  def y_raised_to(i: Int): Double = {
    if(!ypow.isDefinedAt(i))
      ypow += i -> y_raised_to(i-1)*y
    ypow(i)
  }

  def eval(expr: Bn): Double = {
    var answer: Double = 0
    expr.entity.foreach{term =>
      answer += term._2 * x_raised_to(term._1._1) * y_raised_to(term._1._2)
    }
    answer
  }

  def evalAll(list: List[Bn]): List[Double] = {
    list.map(eval)
  }
}

case class Bn(var entity: Map[(Int, Int), Double]) {
  this.reduce

  def this(x: Double) = this(Map((0,0)->x))

  def this(x: Int, y: Int, v: Double) = this(Map((x, y)->v))


  def reduce: Bn = {
    entity = entity.filter(_._2 != 0)
    if(entity.isEmpty) entity = Map((0,0)->0)
    this
  }

  def +(that: Bn): Bn = {
    var temp: Map[(Int, Int), Double] = Map()
    entity.foreach(temp+=)
    that.entity.foreach(tpl =>
      if(temp.isDefinedAt(tpl._1))
        temp(tpl._1) += tpl._2
      else
        temp += tpl
    )
    (new Bn(temp)).reduce
  }

  def +(x: Double): Bn = {
    (this + new Bn(x)).reduce
  }

  def -(that: Bn): Bn = {
    var temp: Map[(Int, Int), Double] = Map()
    entity.foreach(temp+=)
    that.entity.foreach(tpl =>
      if(temp.isDefinedAt(tpl._1))
        temp(tpl._1) -= tpl._2
      else
        temp += tpl._1 -> -tpl._2
    )
    (new Bn(temp)).reduce
  }

  def -(x: Double): Bn = {
    (this - new Bn(x)).reduce
  }


  def *(that: Bn): Bn = {
    var answer = new Bn(0)
    for(x <- entity; y <- that.entity)
      answer += new Bn(Map((x._1._1 + y._1._1, x._1._2 + y._1._2) -> x._2 * y._2))
    answer.reduce
  }

  def *(x: Double): Bn = {
    (this * new Bn(x)).reduce
  }

  def +=(that: Bn) = {
    that.entity.foreach(tpl =>
      if(entity.isDefinedAt(tpl._1))
        entity(tpl._1) += tpl._2
      else
        entity += tpl
    )
    this.reduce
  }

  def -=(that: Bn) = {
    that.entity.foreach(tpl =>
      if(entity.isDefinedAt(tpl._1))
        entity(tpl._1) -= tpl._2
      else
        entity += tpl._1 -> -tpl._2
    )
    this.reduce
  }

  def *=(that: Bn) = {
    var temp = new Bn(0)
    for(x <- entity; y <- that.entity)
      temp += new Bn(Map((x._1._1 + y._1._1, x._1._2 + y._1._2) -> x._2 * y._2))
    entity = temp.entity
    this.reduce
  }

  override def toString(): String = {
    var s: String = ""
    var terms: List[(Int, Int, Double)] = Nil
    entity.map(tpl => terms ::= (tpl._1._1, tpl._1._2, tpl._2))
    terms = terms.sort((x,y)=> x._1<y._1 || x._1==y._1 && x._2<y._2).reverse
    terms.foreach{tpl =>
      if(s == "" && tpl._3 > 0)
        s += tpl._3
      else if(tpl._3 > 0)
        s += "+" + tpl._3
      else
        s += "-" + -tpl._3
      tpl._1 match{
        case 0 =>
        case 1 => s += "x"
        case i => s += "x^" + i
      }
      tpl._2 match{
        case 0 =>
        case 1 => s += "y"
        case j => s += "y^" + j
      }
    }
    s
  }
}






























