/**
 * Created by IntelliJ IDEA.
 * User: Kilho Shin
 * Date: 11/02/21
 * Time: 11:43
 * To change this template use File | Settings | File Templates.

 This file is part of Mapping Kernels for Trees.

 Mapping Kernels for Trees is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Mapping Kernels for Trees is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Mapping Kernels for Trees.  If not, see <http://www.gnu.org/licenses/>.

 For more details about this code, please look at the companion paper
 "Mapping Kernels for Trees", K. Shin, M.Cuturi, T.Kuboyama, Proceedings of
 the International Conference on Machine Learning 2011.

 */

import scala.actors._
import scala.actors.Actor._
import scala.collection.mutable.Map
//import scala.collection.mutable.Set
import scala.collection.mutable.ListBuffer
// import java.io.FileOutputStream
import java.io.OutputStreamWriter

case class Calculate(
  x: (Int, Set[RefNode], Set[RefNode]),
  y: (Int, Set[RefNode], Set[RefNode]))


class DivCompManager(
  val orgzr: Organizer,
  val db: Map[Int, Option[Vertex]], //なんでいるんだっけ？db.sizeを与えれば十分？
  val str: List[(Int, Set[RefNode])],
  val krnList: List[(RefNode, RefNode)=>Bn],
  val fncName: Map[(RefNode, RefNode)=>Bn, String],
  val eval: ((RefNode, RefNode)=>Bn, RefNode, RefNode) => Bn,
  val eval_db: Map[(Signature, Signature), Map[((RefNode, RefNode)=>Bn, RefNode, RefNode), Bn]],
  val parallel_n: Int,
  val blocks: List[(List[Int], List[(Int, Set[RefNode])])],
  //val tempOut: Map[(RefNode, RefNode)=>Bn, OutputStreamWriter],　並列出力
  val krnlWrite: List[(RefNode, RefNode)=>Bn],
  val out: OutputStreamWriter,
  val log: Option[OutputStreamWriter],
  val debug: Boolean,
  val quiet: Boolean) extends Actor {

  var buffer: List[(((RefNode, RefNode)=>Bn, RefNode, RefNode), Bn)] = Nil
  var calculators: List[CellCalculator] = Nil
  1 to parallel_n foreach{_ =>
    val calculator = new CellCalculator(this, krnList)
    calculator.start
    calculators ::= calculator
  }
  var count = parallel_n
  /* 並列出力
  var writers: List[KernelWriter] = Nil
  krnList.foreach {k =>
      val writer = new KernelWriter(this, db, k, fn(k), tempOut(k), eval)
      writers ::= writer
      writer.start
    }
  */

/*
  val data_size: Int = db.size
  val block_n: Int = (data_size + block_size - 1)/block_size
  val blocks =
    1.until(block_n).map(i => ((i-1)*block_size).until(i*block_size).toSet).toList :::
    List(((block_n-1)*block_size).until(data_size).toSet)
*/
  var xrngIt = blocks.iterator
  var xrange = xrngIt.next
  var yrngIt = blocks.iterator
  var yrange = yrngIt.next
  var xscope: List[(Int, Set[RefNode], Set[RefNode])] = null
  var yscope: List[(Int, Set[RefNode], Set[RefNode])] = null

  var cell_n: Int = 0
  var cell_count: Int = 0

  //xscope, yscopeは、カーネル計算の引数となるべき部分構造の集合をBox毎に与える。
  //xscopeは第一引数、yscopeは第二引数に与えられる部分構造。
  //同じサイズの部分構造は一つのエントリーにまとめられる。
  //xscopeとyscopeとに共通で現れる部分構造については、カーネル関数の対称性を
  //利用して重複計算を避けることができる。
  //xscope, yscopeの各エントリーの第一要素は部分構造のサイズ(Int)、
  //第二要素は共通に含まれる部分構造の集合(Set[RefNode])、
  //第三要素は共通でない部分構造の集合(Set[RefNode])を表す
  var x: (Int, Set[RefNode], Set[RefNode]) = null
  var y: (Int, Set[RefNode], Set[RefNode]) = null
  var xIt: Iterator[(Int, Set[RefNode], Set[RefNode])] = null
  var yIt: Iterator[(Int, Set[RefNode], Set[RefNode])] = null
/*
  print("*** Total "+data_size+" data are partitioned into intervals ")
  blocks.init.foreach(x=>print("["+x.min+".."+x.max+"], "))
  println("and ["+blocks.last.min+".."+blocks.last.max+"].")
*/
  determine_xscope_and_yscope
  move_to_next_cell

  def determine_xscope_and_yscope {
/*
    var xtmp: Set[RefNode] = Set()
    var ytmp: Set[RefNode] = Set()
    xscope = Nil
    yscope = Nil
    str.foreach{tpl =>
      val xtmp = tpl._2.filter(!_.referencedFrom.intersect(xrange).isEmpty)
      val ytmp = tpl._2.filter(!_.referencedFrom.intersect(yrange).isEmpty)
      val common = xtmp.intersect(ytmp)
      if(!xtmp.isEmpty)
        xscope ::= (tpl._1, common, xtmp -- common)
      if(!ytmp.isEmpty)
        yscope ::= (tpl._1, common, ytmp -- common)
    }*/
   xscope = Nil
   yscope = Nil
   xrange._2.zip(yrange._2).foreach{
     tpl =>
     val xtmp: Set[RefNode] = tpl._1._2
     val ytmp: Set[RefNode] = tpl._2._2
     val common = xtmp.intersect(ytmp)
     if(!xtmp.isEmpty)
       xscope ::= (tpl._1._1, common, xtmp -- common)
     if(!ytmp.isEmpty)
       yscope ::= (tpl._1._1, common, ytmp -- common)
   }
   xscope = xscope.reverse
   yscope = yscope.reverse

    xIt = xscope.iterator
    x = xIt.next
    yIt = yscope.iterator
    y = yIt.next

    if(!quiet) {
      cell_n = xscope.size * yscope.size
      cell_count = 0
      val xcn = xscope.foldLeft(0)(_+_._2.size)
      val ycn = yscope.foldLeft(0)(_+_._2.size)
      val xpn = xscope.foldLeft(0)(_+_._3.size)
      val ypn = yscope.foldLeft(0)(_+_._3.size)
      print("*** Calculating kernel values for [")
      print(xrange._1.min+".."+xrange._1.max+"]("+xcn+"+"+xpn+"="+(xcn+xpn)+") x ")
      println("["+yrange._1.min + ".."+yrange._1.max+"]("+ycn+"+"+ypn+"="+(ycn+ypn)+")")
      print("***** Progress: # of cells completed / " + cell_n + " : ")
    }
  }

  def move_to_next_cell {
    if(!quiet) {
      cell_count += 1
      if(cell_count % 10 == 0) print(cell_count+" ")
    }

    val cn: Int = x._2.size/parallel_n
    var ctmp = x._2
    val pn: Int = x._3.size/parallel_n
    var ptmp = x._3
    calculators.init.foreach{calculator =>
      val (x2, tmp2) = ctmp.splitAt(cn)
      ctmp = tmp2
      val (x3, tmp3) = ptmp.splitAt(pn)
      ptmp = tmp3
      //println("********* Partitioned size  = " + x2.size + " : " + x3.size + " , " + y._2.size + " : " + y._3.size)
      calculator ! Calculate((x._1, x2, x3), y)
    }
    //println("********* Partitioned size  = " + ctmp.size + " : " + ptmp.size + " , " + y._2.size + " : " + y._3.size)
    calculators.last ! Calculate((x._1, ctmp, ptmp), y)
    count = parallel_n
  }

  def act {
    loop {
      react {
        case Return(p: Actor, r: List[(((RefNode, RefNode)=>Bn, RefNode, RefNode), Bn)]) =>
          //println("********* Receive for "+x._1+"x"+y._1+" count = "+count)
          store_returned_result_to_buffer(r)
          if(current_cell_not_finished_yet)
            wait_for_remaining_results_from_cell_calculators
          else {
            store_buffered_data_to_database_then_clear_buffer
            if(next_cell_exists_in_current_box)
              move_to_next_cell
            else {
              write_kernel_values_of_current_box
              if(next_box_exists)
                move_to_first_cell_of_next_box
              else
                complete_mission
            }
          }
      }
    }
  }

  def store_returned_result_to_buffer(r: List[(((RefNode, RefNode)=>Bn, RefNode, RefNode), Bn)]) {
    count -= 1
    buffer :::= r
  }

  def current_cell_not_finished_yet: Boolean = {
    count > 0
  }

  def wait_for_remaining_results_from_cell_calculators {}

  def store_buffered_data_to_database_then_clear_buffer {
    buffer.foreach{tpl =>
      var r1 = tpl._1._2
      var r2 = tpl._1._3
      if(r2.signature < r1.signature){
        val tmp = r1
        r1 = r2
        r2 = tmp
      }
      var s1 = r1.signature
      var s2 = r2.signature
      if(!eval_db.isDefinedAt(s1, s2))
        eval_db += (s1, s2) -> Map()
      eval_db(s1, s2) += (tpl._1._1, r1, r2)->tpl._2
    }
    buffer = Nil
    //print(x._1+"x"+y._1+"->"+(result.size/krnList.length)+",")
  }

  def next_cell_exists_in_current_box: Boolean = {
    if(yIt.hasNext) {
      y = yIt.next
      true
    }
    else if(xIt.hasNext) {
      x = xIt.next
      yIt = yscope.iterator
      y = yIt.next
      true
    } else
      false
  }

  def next_box_exists: Boolean = {
    if(yrngIt.hasNext) {
      yrange = yrngIt.next
      determine_xscope_and_yscope
      true
    } else if(xrngIt.hasNext) {
      xrange = xrngIt.next
      yrngIt = blocks.dropWhile(xrange!=).iterator
      yrange = yrngIt.next
      determine_xscope_and_yscope
      true
    } else
      false
  }

  def writing_finishes: Boolean = {
    count -= 1
    count == 0
  }

  def write_kernel_values_of_current_box {
    if(!quiet){
      val db_size = eval_db.foldLeft(0)(_+_._2.size)
      println("\n******* Total number of kernel values calculated = " + db_size)
      println("***** Outputing")
    }
    for(k<-krnlWrite) {
      out.write("*** "+fncName(k)+": ["+xrange._1.min+","+xrange._1.max+"]x["+yrange._1.min+","+yrange._1.max+"]\r\n")
      for(i <- xrange._1.sort(_<_)){
        var s = i+": "
        for(j <- yrange._1.sort(_<_)) {
          val r1 = db(i).get.children.head.refAsTree
          val r2 = db(j).get.children.head.refAsTree
          var v: Bn = null
          s += j+":"+eval(k, r1, r2)+" "
        }
        out.write(s+"\r\n")
      }
    }
    eval_db.keySet.foreach(eval_db.remove)
    System.gc
  }

  def move_to_first_cell_of_next_box {
    move_to_next_cell
  }

  def complete_mission {
    orgzr ! Finish(this)
    calculators.foreach(_ ! Exit)
    // writers.foreach(_ ! Exit)
    exit()
  }
}
